@@ -26,7 +26,7 @@
  *
  */
 char radrelay_rcsid[] =
-"$Id: radrelay.c,v 1.22.2.3.2.3 2007/03/05 14:26:11 aland Exp $";
+"$Id: radrelay.c,v 1.22.2.3.2.4 2007/03/16 13:22:03 aland Exp $";
 
 #include "autoconf.h"
 
@@ -389,9 +389,10 @@ int do_recv(struct relay_misc *r_args)
 	 *	Must be an accounting response.
 	 *	FIXME: check if this is the right server!
 	 */
-	if (rep->code != PW_ACCOUNTING_RESPONSE)
+	if (rep->code != PW_ACCOUNTING_RESPONSE) {
 		rad_free(&rep);
 		return -1;
+	}
 
 	/*
 	 *	Decode packet into radius attributes.
