@@ -1,7 +1,7 @@
 /*
  * eap_tls.c
  *
- * Version:     $Id: eap_tls.c,v 1.1.2.3 2007/03/05 14:11:55 aland Exp $
+ * Version:     $Id: eap_tls.c,v 1.1.2.4 2007/04/07 21:27:19 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -374,7 +374,8 @@ static eaptls_status_t eaptls_verify(EAP_HANDLER *handler)
 			 * 2. If EAP-response received, then its M bit not set.
 			 * 	(It is because Last fragment will not have M bit set)
 			 */
-			if ((prev_eap_ds->response == NULL) ||
+			if (!prev_eap_ds ||
+			    (prev_eap_ds->response == NULL) ||
 			    (eaptls_prev == NULL) ||
 			    !TLS_MORE_FRAGMENTS(eaptls_prev->flags)) {
 
