@@ -1,7 +1,7 @@
 /*
  * rlm_pap.c
  *
- * Version:  $Id: rlm_pap.c,v 1.15.2.2.2.4 2007/03/05 13:12:29 aland Exp $
+ * Version:  $Id: rlm_pap.c,v 1.15.2.2.2.5 2007/04/08 06:31:38 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -49,7 +49,7 @@
 #define PAP_MAX_ENC		9
 
 
-static const char rcsid[] = "$Id: rlm_pap.c,v 1.15.2.2.2.4 2007/03/05 13:12:29 aland Exp $";
+static const char rcsid[] = "$Id: rlm_pap.c,v 1.15.2.2.2.5 2007/04/08 06:31:38 aland Exp $";
 
 /*
  *      Define a structure for our module configuration.
@@ -583,7 +583,7 @@ static int pap_authenticate(void *instance, REQUEST *request)
 			goto make_msg;
 		}
 	done:
-		DEBUG("rlm_pap: User authenticated succesfully");
+		DEBUG("rlm_pap: User authenticated successfully");
 		return RLM_MODULE_OK;
 		break;
 		
