@@ -1,7 +1,7 @@
 /*
- * rlm_radutmp.c	
+ * rlm_radutmp.c
  *
- * Version:	$Id: rlm_radutmp.c,v 1.19.2.1 2003/07/11 14:12:44 aland Exp $
+ * Version:	$Id: rlm_radutmp.c,v 1.24.4.1 2007/04/07 22:40:00 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -143,7 +143,7 @@ static int radutmp_zap(rlm_radutmp_t *inst,
 	 *	Lock the utmp file, prefer lockf() over flock().
 	 */
 	rad_lockfd(fd, LOCK_LEN);
-	
+
 	/*
 	 *	Find the entry for this NAS / portno combination.
 	 */
@@ -302,7 +302,7 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 					sizeof(ut.session_id));
 				break;
 			case PW_NAS_PORT_TYPE:
-				if (vp->lvalue >= 0 && vp->lvalue <= 4)
+				if (vp->lvalue <= 4)
 					ut.porttype = porttypes[vp->lvalue];
 				nas_port_type = vp->lvalue;
 				break;
@@ -441,7 +441,7 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 	 *	Lock the utmp file, prefer lockf() over flock().
 	 */
 	rad_lockfd(fd, LOCK_LEN);
-	
+
 	/*
 	 *	Find the entry for this NAS / portno combination.
 	 */
@@ -449,7 +449,7 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 				   ut.nas_port)) != NULL) {
 		lseek(fd, (off_t)cache->offset, SEEK_SET);
 	}
-	
+
 	r = 0;
 	off = 0;
 	while (read(fd, &u, sizeof(u)) == sizeof(u)) {
@@ -457,7 +457,15 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 		if (u.nas_address != ut.nas_address ||
 		    u.nas_port	  != ut.nas_port)
 			continue;
-		
+
+		/*
+		 *	Don't compare stop records to unused entries.
+		 */
+		if (status == PW_STATUS_STOP &&
+		    u.type == P_IDLE) {
+			continue;
+		}
+
 		if (status == PW_STATUS_STOP &&
 		    strncmp(ut.session_id, u.session_id,
 			    sizeof(u.session_id)) != 0) {
@@ -472,7 +480,7 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 			r = -1;
 			break;
 		}
-		
+
 		if (status == PW_STATUS_START &&
 		    strncmp(ut.session_id, u.session_id,
 			    sizeof(u.session_id)) == 0  &&
@@ -488,7 +496,7 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 			r = -1;
 			break;
 		}
-		
+
 		/*
 		 *	FIXME: the ALIVE record could need
 		 *	some more checking, but anyway I'd
@@ -505,7 +513,7 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 			if (u.login[0] != 0)
 				just_an_update = 1;
 		}
-		
+
 		if (lseek(fd, -(off_t)sizeof(u), SEEK_CUR) < 0) {
 			radlog(L_ERR, "rlm_radutmp: negative lseek!");
 			lseek(fd, (off_t)0, SEEK_SET);
@@ -534,11 +542,11 @@ static int radutmp_accounting(void *instance, REQUEST *request)
 			cache->next = inst->nas_port_list;
 			inst->nas_port_list = cache;
 		}
-		
+
 		ut.type = P_LOGIN;
 		write(fd, &ut, sizeof(u));
 	}
-	
+
 	/*
 	 *	The user has logged off, delete the entry by
 	 *	re-writing it in place.
@@ -581,6 +589,7 @@ static int radutmp_checksimul(void *instance, REQUEST *request)
 	rlm_radutmp_t	*inst = instance;
 	char		login[256];
 	char		filename[1024];
+	ssize_t		read_size;
 
 	/*
 	 *	Get the filename, via xlat.
@@ -621,7 +630,7 @@ static int radutmp_checksimul(void *instance, REQUEST *request)
 	/*
 	 *	Loop over utmp, counting how many people MAY be logged in.
 	 */
-	while (read(fd, &u, sizeof(u)) == sizeof(u)) {
+	while ((read_size = read(fd, &u, sizeof(u))) == sizeof(u)) {
 		if (((strncmp(login, u.login, RUT_NAMESIZE) == 0) ||
 		     (!inst->case_sensitive &&
 		      (strncasecmp(login, u.login, RUT_NAMESIZE) == 0))) &&
@@ -630,6 +639,13 @@ static int radutmp_checksimul(void *instance, REQUEST *request)
 		}
 	}
 
+	if (read_size < 0) {
+		radlog(L_ERR, "rlm_radutmp: Error reading %s: %s",
+		       filename, strerror(errno));
+		close(fd);
+		return RLM_MODULE_FAIL;
+	}
+
 	/*
 	 *	The number of users logged in is OK,
 	 *	OR, we've been told to not check the NAS.
@@ -647,7 +663,7 @@ static int radutmp_checksimul(void *instance, REQUEST *request)
 	if ((vp = pairfind(request->packet->vps, PW_FRAMED_IP_ADDRESS)) != NULL)
 		ipno = vp->lvalue;
 	if ((vp = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) != NULL)
-		call_num = vp->strvalue;	
+		call_num = vp->strvalue;
 
 	/*
 	 *	lock the file while reading/writing.
@@ -661,7 +677,7 @@ static int radutmp_checksimul(void *instance, REQUEST *request)
 	 *	static IP's like DSL.
 	 */
 	request->simul_count = 0;
-	while (read(fd, &u, sizeof(u)) == sizeof(u)) {
+	while ((read_size = read(fd, &u, sizeof(u))) == sizeof(u)) {
 		if (((strncmp(login, u.login, RUT_NAMESIZE) == 0) ||
 		     (!inst->case_sensitive &&
 		      (strncasecmp(login, u.login, RUT_NAMESIZE) == 0))) &&
@@ -698,7 +714,7 @@ static int radutmp_checksimul(void *instance, REQUEST *request)
 			rcode = rad_check_ts(u.nas_address, u.nas_port,
 					     utmp_login, session_id);
 			rad_lockfd(fd, LOCK_LEN);
-			
+
 			/*
 			 *	Failed to check the terminal server for
 			 *	duplicate logins: Return an error.
@@ -725,13 +741,21 @@ static int radutmp_checksimul(void *instance, REQUEST *request)
 				/*
 				 *	False record - zap it.
 				 */
-				session_zap(request->packet->sockfd,
+				session_zap(request,
 					    u.nas_address, u.nas_port, login,
 					    session_id, u.framed_address,
-					    u.proto, 0);
+					    u.proto);
 			}
 		}
 	}
+
+	if (read_size < 0) {
+		radlog(L_ERR, "rlm_radutmp: Error reading %s: %s",
+		       filename, strerror(errno));
+		close(fd);
+		return RLM_MODULE_FAIL;
+	}
+
 	close(fd);		/* and implicitely release the locks */
 
 	return RLM_MODULE_OK;
